#!/usr/bin/env python3
"""This module provides functions to access RedPen server.

This module may use the following environment variables.

:SABACAN_REDPEN_URL:
    URL of RedPen server. If SABACAN_REDPEN_URL does not exist,
    use SABACAN_URL instead.
:SABACAN_REDPEN_TIMEOUT:
    Timeout (sec) of RedPen server communication. If SABACAN_REDPEN_TIMEOUT
    does not exist, use SABACAN_TIMEOUT instead.
"""
import argparse
import json
import logging
import os
import pathlib
import re
import sys
import urllib.error
import urllib.parse
import urllib.request
try:
    from lxml import etree as ET
except ImportError:
    from xml.etree import ElementTree as ET

import sabacan.utils
from sabacan.utils import NotSupportedAction


_SERVER_HOST = '127.0.0.1'
_SERVER_PORT = 8080
_DEFAULT_SERVER_URL = 'http://%s:%d' % (_SERVER_HOST, _SERVER_PORT)
_PARSER_EXTENSIONS_LIST = [
        ('markdown', ['md', 'markdown']),
        ('plain', ['txt']),
        ('wiki', ['wiki']),
        ('asciidoc', ['adoc', 'asciidoc']),
        ('latex', ['tex', 'latex']),
        ('rest', ['rest', 'rst']),
        ('review', ['re', 'review']),
        ('properties', ['properties']),
]
_FORMAT_LIST = ['json', 'json2', 'plain', 'plain2', 'xml']


def make_parser(parser_constructor=argparse.ArgumentParser):
    """Make argparse parser object for RedPen server.
    """
    parser = parser_constructor(
        'redpen',
        usage='%(prog)s [Options] [<INPUT FILE>]',
        description='CLI for RedPen server',
        add_help=False)
    parser.add_argument(
        '--conf', '-c',
        help='Configuration file',
        action='store',
        metavar='<CONF FILE>')
    parser.add_argument(
        '--format', '-f',
        help='Input file format (%(choices)s)',
        action='store',
        choices=[e[0] for e in _PARSER_EXTENSIONS_LIST],
        metavar='<FORMAT>',
        dest='document_parser')
    parser.add_argument(
        '--help', '-h',
        help='Displays this help information and exits',
        action='help')
    parser.add_argument(
        '--limit', '-l',
        help='Error limit number (default: %(default)d)',
        action='store',
        type=int,
        metavar='<LIMIT NUMBER>',
        default=1)
    parser.add_argument(
        '--lang', '-L',
        help='Language of error messages (%(choices)s)',
        action='store',
        choices=['en', 'ja'],
        metavar='<LANGUAGE>')
    parser.add_argument(
        '--result-format', '-r',
        help='Output result format (%(choices)s) (default: %(default)s)',
        action='store',
        choices=_FORMAT_LIST,
        metavar='<RESULT FORMAT>',
        dest='format',
        default='plain')
    parser.add_argument(
        '--sentence', '-s',
        help='Input sentences',
        action='store',
        metavar='<INPUT SENTENCES>',
        dest='document')
    parser.add_argument(
        '--threshold', '-t',
        help='Threshold of error level (info, warn, error)',
        action=NotSupportedAction,
        choices=['info', 'warn', 'error'],
        metavar='<THRESHOLD>')
    parser.add_argument(
        '--version', '-v',
        action='store_true',
        help='Displays version information and exits')
    parser.add_argument(
        'input_files',
        help='Input document',
        nargs='*',
        metavar='<INPUT FILE>')
    parser.set_defaults(main_function=main)
    return parser


def get_default_configfile(lang):
    """Get default RedPen XML configuration file.

    Search current directory.
    If no configurations are found, search the parent directories
    until root directory.
    If no configurations are found, search $REDPEN_HOME/conf/.

    Args:
        lang: The languange which searched configuration file name includes.
    Returns:
        pathlib.Path:
            The path to configuration file. If not found, return None.
    """
    redpen_conf_lang_name = 'redpen-conf-' + lang + '.xml'
    def search_conf(dirpath):
        conf = dirpath / 'redpen-conf.xml'
        if conf.is_file():
            return conf
        conf = dirpath / redpen_conf_lang_name
        if conf.is_file():
            return conf
        return None

    current = pathlib.Path('.')
    conf = search_conf(current)
    if conf is not None:
        return conf
    for dirpath in current.resolve().parents:
        conf = search_conf(dirpath)
        if conf is not None:
            return conf
    redpen_home = pathlib.Path(os.getenv('REDPEN_HOME', '.'))
    return search_conf(redpen_home / 'conf')


def get_document_parser_from_filename(filename):
    """Get document format from the file extension.

    Args:
        filename: The filename of document.
    Returns:
        str: The format of the document.
    """
    suffix = pathlib.Path(filename).suffix
    if not suffix:
        return None
    for parser, extensions in _PARSER_EXTENSIONS_LIST:
        if suffix[1:] in extensions:
            return parser
    return None


def get_number_of_errors(result, output_format):
    """Get the number of errors in validation result.

    Args:
        result (str): A validation result.
        output_format (str): The format of the validation result.
    Returns:
        int: The number of errors in the validation result.
    """
    if output_format == 'json':
        result = json.loads(result)
        return len(result[0]['errors'])
    if output_format == 'json2':
        result = json.loads(result)
        return sum((len(error['errors']) for error in result[0]['errors']))
    if output_format == 'plain':
        return len(result.splitlines())
    if output_format == 'plain2':
        errors = re.finditer(r'^\s+(?!(Line|Sentence):)(?=\w)',
                             result, flags=re.MULTILINE)
        return sum((1 for _ in errors))
    if output_format == 'xml':
        root = ET.fromstring(result)
        return len(root.findall('error'))
    return 0 # Unknown format


def get_version(base_url, timeout=None):
    """Get RedPen version.

    Args:
        base_url (str): URL of RedPen server.
        timeout (int): The server communication timeout in seconds.
    Returns:
        str: RedPen version.
    """
    url = base_url + '/rest/config/redpens'
    with urllib.request.urlopen(url, timeout=timeout) as response:
        result = json.loads(response.read().decode('utf8'))
        return result['version']


def get_language(base_url, document, timeout=None):
    """Get language of the document.

    Args:
        base_url (str): URL of RedPen server.
        document (str): Document to get which language uses.
        timeout (int): The server communication timeout in seconds.
    Returns:
        str: The language of the document.
    """
    url = base_url + '/rest/document/language'
    data = {'document': document}
    data = urllib.parse.urlencode(data).encode('utf8')
    with urllib.request.urlopen(url, data, timeout=timeout) as response:
        result = json.loads(response.read().decode('utf8'))
        return result['key']


def validate(base_url, document, document_parser, lang, output_format,
             config=None, timeout=None):
    """Validate document.

    Args:
        base_url (str): URL of RedPen server.
        document (str): Document to be validated.
        document_parser (str): Document format.
        lang (str): The language of document.
        output_format (str): The format of the validation result.
        config (str): The RedPen XML configuration.
        timeout (int): The server communication timeout in seconds.
    Returns:
        str: The validation result with the specified format.
    """
    # pylint: disable=too-many-arguments
    data = {
        'document': document,
        'documentParser': document_parser,
        'lang': lang,
        'format': output_format,
    }
    if config is not None:
        data['config'] = config

    url = base_url + '/rest/document/validate'
    data = urllib.parse.urlencode(data).encode('utf8')
    with urllib.request.urlopen(url, data, timeout=timeout) as response:
        result = response.read().decode('utf8')
        if output_format.startswith('json'):
            return '[%s]' % result
        return result


def _exit_by_error(msg, *args, **kwargs):
    logging.error(msg, *args, **kwargs)
    sys.exit(1)


def _get_config(args):
    logging.debug('Getting configuration file...')
    if args.conf is None:
        return None
    config_file = pathlib.Path(args.conf)
    if not config_file.is_file():
        _exit_by_error('%s is not file', config_file)
    return config_file.read_text()

def _get_lang_from_config(config):
    logging.debug('Getting language from configuration file...')
    root = ET.fromstring(config)
    lang = root.attrib.get('lang', 'en')
    return lang

def _get_document_parser(args, document):
    if args.document_parser is not None:
        return args.document_parser
    if document.filename is not None:
        parser = get_document_parser_from_filename(document.filename)
        if parser is not None:
            return parser
    return 'plain'

def _get_default_config(lang, config_cache):
    if lang in config_cache:
        return config_cache[lang]
    logging.debug('Getting %s configuration file...', lang)
    config_file = get_default_configfile(lang)
    if config_file is None:
        logging.debug('Not found %s configuration file...', lang)
        config = None
    else:
        logging.debug('Found configuration file: %s', config_file)
        config = config_file.read_text()
    config_cache[lang] = config
    return config


class _Document:
    def __init__(self, document):
        self._is_file = isinstance(document, pathlib.Path)
        self._document = document

    @property
    def filename(self):
        """Get filename"""
        if self._is_file:
            return self._document.name
        return None

    @property
    def document(self):
        """Get document"""
        if self._is_file:
            return self._document.read_text()
        return self._document

def _get_documents(args):
    logging.debug('Getting documents...')
    if args.document is not None:
        return [_Document(args.document)]
    if not args.input_files:
        _exit_by_error('Input is not given')
    docs = []
    for input_file in args.input_files:
        path = pathlib.Path(input_file)
        if not path.is_file():
            _exit_by_error('Input is not regular file')
        docs.append(_Document(path))
    return docs

class _PlainMerger:
    def __init__(self):
        self._result = []

    def __call__(self, name, result):
        lines = result.splitlines()
        if name is not None:
            self._result.extend(name + ':' + line for line in lines)
        else:
            self._result.extend(lines)

    def __str__(self):
        return '\n'.join(self._result)

class _Plain2Merger:
    def __init__(self):
        self._result = ''

    def __call__(self, name, result):
        if name is not None:
            self._result += 'Document: ' + name + '\n'
        self._result += result

    def __str__(self):
        return self._result

class _JSONMerger:
    def __init__(self):
        self._result = []

    def __call__(self, name, result):
        json_result = json.loads(result)
        if name is not None:
            json_result[0]['document'] = name
        self._result.append(json_result[0])

    def __str__(self):
        return json.dumps(self._result, ensure_ascii=False)

class _XMLMerger:
    def __init__(self):
        self._result = ET.Element('validation-result')

    def __call__(self, name, result):
        root = ET.fromstring(result)
        for error in root.iterfind('error'):
            if name is not None:
                file_elem = ET.Element('file')
                file_elem.text = name
                idx = len(error)
                for idx, elem in enumerate(error):
                    if elem.tag == 'lineNum':
                        break
                error.insert(idx, file_elem)
            self._result.append(error)

    def __str__(self):
        return ET.tostring(self._result, encoding='utf8').decode('utf8')

def _merge_result(results, output_format):
    merger_map = {
        'plain': _PlainMerger,
        'plain2': _Plain2Merger,
        'json': _JSONMerger,
        'json2': _JSONMerger,
        'xml': _XMLMerger,
    }
    merger = merger_map[output_format]()
    for name, result in results:
        merger(name, result)
    return str(merger)


def main(args):
    """Run action as redpen command.

    Args:
        args: Parsing result from the parser created by `make_parser`.
    """
    base_url = sabacan.utils.get_server_url('redpen', _DEFAULT_SERVER_URL)
    timeout = sabacan.utils.get_timeout('redpen')

    if args.version:
        logging.debug('Getting RedPen version...')
        print(get_version(base_url, timeout=timeout))
        sys.exit(0)

    global_config = _get_config(args)
    if global_config is not None:
        lang = _get_lang_from_config(global_config)
    config_cache = {}
    results = []
    num_error = 0
    for doc in _get_documents(args):
        logging.debug('Getting document parser...')
        document_parser = _get_document_parser(args, doc)
        contents = doc.document
        if global_config is None:
            if args.lang is None:
                logging.debug('Getting language from input document...')
                lang = get_language(base_url, contents, timeout=timeout)
            else:
                lang = args.lang
            config = _get_default_config(lang, config_cache)
        else:
            config = global_config

        logging.debug('Validating input document '
                      '(document_parser=%s, lang=%s, format=%s)...',
                      document_parser, lang, args.format)
        try:
            result = validate(base_url,
                              contents, document_parser, lang, args.format,
                              config=config, timeout=timeout)
        except urllib.error.HTTPError as error:
            with error:
                _exit_by_error('Failed to validate input document (%d %s): %s',
                               error.code, error.reason, error.read())
        results.append((doc.filename, result))

        logging.debug('Calculating the number of errors...')
        num_error += get_number_of_errors(result, args.format)

    print(_merge_result(results, args.format))
    logging.debug('The number of errors is %d', num_error)
    if args.limit < num_error:
        _exit_by_error('The number of errors "%d" is larger'
                       ' than specified (limit is "%d")',
                       num_error, args.limit)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logging.debug('Parsing command line options...')
    ARGS = make_parser().parse_args()
    ARGS.main_function(ARGS)



"""This module provides sabacan command main function.

This module may use the following environment variables.

:SABACAN_URL:
    URL of server.
:SABACAN_TIMEOUT:
    Timeout (sec) of server communication.
"""
import argparse
import sabacan.plantuml
import sabacan.redpen
from sabacan.utils import SetEnvAction


__version__ = '0.0.1'


def main():
    """Run sabacan command.
    """
    parser = argparse.ArgumentParser('sabacan')
    parser.add_argument(
        '--version', '-v',
        action='version',
        version='%%(prog)s %s' % __version__)
    parser.add_argument(
        '--url', '-u',
        help='server URL',
        metavar='URL',
        action=SetEnvAction,
        dest='SABACAN_URL')
    parser.add_argument(
        '--timeout', '-t',
        help='server communication timeout (sec)',
        metavar='N',
        type=int,
        action=SetEnvAction,
        dest='SABACAN_TIMEOUT')

    subparsers = parser.add_subparsers(
        title='supported subcommand',
        help='%(choices)s',
        metavar='subcommand')
    sabacan.plantuml.make_parser(subparsers.add_parser)
    sabacan.redpen.make_parser(subparsers.add_parser)

    args = parser.parse_args()
    args.main_function(args)



#!/usr/bin/env python3
"""This module provides functions to access PlantUML server.

This module may use the following environment variables.

:SABACAN_PLANTUML_URL:
    URL of RedPen server. If SABACAN_PLANTUML_URL does not exist,
    use SABACAN_URL instead.
:SABACAN_PLANTUML_TIMEOUT:
    Timeout (sec) of RedPen server communication. If SABACAN_PLANTUML_TIMEOUT
    does not exist, use SABACAN_TIMEOUT instead.
"""
import argparse
import base64
import glob
import logging
import os
import pathlib
import sys
import urllib.error
import urllib.request
import zlib

import sabacan.utils
from sabacan.utils import NotSupportedAction, NotSupportedFlagAction

_FROM_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/='
_TO_CHARS = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz-_?'
_ENCODE_TABLE = dict(zip(_FROM_CHARS, _TO_CHARS))
_DECODE_TABLE = dict(zip(_TO_CHARS, _FROM_CHARS))

_FORMAT_LIST = [
    'png', 'braille',
    'svg', 'svg:nornd',
    'eps', 'eps:text',
    'pdf', 'vdx',
    'xmi', 'xmi:argo', 'xmi:star',
    'scxml', 'html',
    'txt', 'utxt',
    'latex', 'latex:nopreamble'
    'base64',
]
_FORMAT_TO_URL_PATTERN_TABLE = {
    'eps:text': 'epstext',
    'utxt': 'txt',
}

DEFAULT_SERVER_URL = 'http://%s:%d/plantuml' % ('127.0.0.1', 8080)


class _FlagAction(argparse.Action):
    """Custom argparse.Action class to display variable information.

    This class compile dest as PlantUML code when the action is invoked.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        # pylint: disable=too-many-arguments,redefined-builtin
        super(_FlagAction, self).__init__(option_strings, dest, 0, const,
                                         default, type, choices, required,
                                         help, metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        base_url = sabacan.utils.get_server_url('plantuml', DEFAULT_SERVER_URL)
        timeout = sabacan.utils.get_timeout('plantuml')
        code = '@startuml\n' + self.dest + '\n@enduml'
        try:
            result = compile_code(base_url, code, 'txt', timeout=timeout)
            lines = result.decode('utf-8').splitlines()
            print('\n'.join(line.strip() for line in lines))
        except Exception as ex: # pylint: disable=broad-except
            logging.error('Failed to display %s: %s', self.dest, ex)
            parser.exit(1)
        parser.exit(0)


class _LanguageAction(argparse.Action):
    """Custom argparse.Action class to display PlantUML language information.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        # pylint: disable=too-many-arguments,redefined-builtin
        super(_LanguageAction, self).__init__(
            option_strings, dest, 0, const,
            default, type, choices, required, help, metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        base_url = sabacan.utils.get_server_url('plantuml', DEFAULT_SERVER_URL)
        timeout = sabacan.utils.get_timeout('plantuml')
        try:
            print(get_language(base_url, timeout))
        except Exception as ex: # pylint: disable=broad-except
            logging.error('Failed to display language: %s', ex)
            parser.exit(1)
        parser.exit(0)


def make_parser(parser_constructor=argparse.ArgumentParser):
    """Make argparse parser object for PlantUML server.
    """
    # pylint: disable=too-many-statements
    parser = parser_constructor(
        'plantuml',
        usage='%(prog)s [options] [file/dir] [file/dir] [file/dir]',
        description='CLI for PlantUML server',
        add_help=False)
    parser.add_argument(
        '-gui',
        help='To run the graphical user interface',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-t',
        help='To generate images using specified format (default is PNG)',
        action='store',
        choices=_FORMAT_LIST,
        default='png',
        metavar='format',
        dest='format')
    parser.add_argument(
        '-output', '-o',
        help='To generate images in the specified directory',
        action='store',
        metavar='"dir"',
        dest='outdir')
    parser.add_argument(
        '-ofile',
        help='To generate an image with the specified filename',
        action='store',
        metavar='"file"',
        dest='outfile')
    parser.add_argument(
        '-D',
        action='store',
        metavar='VAR1=value',
        help=('Not support currently. '
              'To set a preprocessing variable '
              'as if \'!define VAR1 value\' were used'))
    parser.add_argument(
        '-S',
        help=('Not support currently. '
              'To set a skin parameter '
              'as if \'skinparam param1 value\' were used'),
        action='store',
        metavar='param1=value')
    parser.add_argument(
        '-recurse', '-r',
        help='recurse through directories',
        action='store_true')
    parser.add_argument(
        '-config', '-c',
        help=('Not support currently. '
              'To read the provided config file before each diagram'),
        metavar='"file"',
        action='store')
    parser.add_argument(
        '-I',
        help=('Not support currently. '
              'To include file as if \'!include file\' were used'),
        action='store',
        metavar='/path/to/file')
    parser.add_argument(
        '-charset',
        help=('Not support currently. '
              'To use a specific charset (default is US-ASCII)'),
        action='store',
        metavar='xxx')
    parser.add_argument(
        '-exclude', '-x',
        help=('Not support currently. '
              'To exclude files that match the provided pattern'),
        action='store',
        metavar='pattern')
    parser.add_argument(
        '-metadata',
        help='To retrieve PlantUML sources from PNG images',
        action='store_true')
    parser.add_argument(
        '-nometadata',
        help='To NOT export metadata in PNG/SVG generated files',
        action='store_true')
    parser.add_argument(
        '-checkmetadata',
        help='Skip PNG files that don\'t need to be regenerated',
        action='store_true')
    parser.add_argument(
        '-logdata',
        help='TODO',
        action='store',
        metavar='"file"')
    parser.add_argument(
        '-word',
        help='TODO',
        action='store_true')
    parser.add_argument(
        '-version',
        help='To display information about PlantUML and Java versions',
        action=_FlagAction)
    parser.add_argument(
        '-license',
        help='To display license',
        action=_FlagAction)
    parser.add_argument(
        '-checkversion',
        help='To check if a newer version is available for download',
        action=_FlagAction)
    parser.add_argument(
        '-verbose', '-v',
        help='To have log information',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-quiet',
        help=('Not support currently. '
              'To NOT print error message into the console'),
        action='store_true')
    parser.add_argument(
        '-debugsvek', '-debug_svek',
        help='To generate intermediate svek files',
        action=NotSupportedFlagAction)
    # parser.add_argument(
    #     '-keepfiles',
    #     help='NOT delete temporary files after process.',
    #     action=NotSupportedFlagAction)
    parser.add_argument(
        '-help', '-h', '-?',
        help='To display this help message',
        action='help')
    parser.add_argument(
        '-testdot',
        help='To test the installation of graphviz',
        action=_FlagAction)
    parser.add_argument(
        '-graphvizdot', '-graphviz_dot',
        help='To specify dot executable',
        action=NotSupportedAction,
        metavar='"exe"')
    parser.add_argument(
        '-pipe', '-p',
        help=('To use stdin for PlantUML source '
              'and stdout for PNG/SVG/EPS generation'),
        action='store_true')
    parser.add_argument(
        '-encodesprite',
        help=('To encode a sprite at gray level (z for compression) '
              'from an image'),
        action=NotSupportedAction,
        choices=['4', '8', '16', '4z', '8z', '16z'])
    parser.add_argument(
        '-computeurl', '-encodeurl',
        help='To compute the encoded URL of a PlantUML source file',
        action='store_true')
    parser.add_argument(
        '-decodeurl',
        help='To retrieve the PlantUML source from an encoded URL',
        action='store_true')
    parser.add_argument(
        '-syntax',
        help=('To report any syntax error '
              'from standard input without generating images'),
        action='store_true')
    parser.add_argument(
        '-language',
        help='To print the list of PlantUML keywords',
        action=_LanguageAction)
    # parser.add_argument(
    #     '-nosuggestengine',
    #     action='store_true',
    #     help='To disable the suggest engine when errors in diagrams')
    parser.add_argument(
        '-checkonly',
        help='To check the syntax of files without generating images',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-failfast',
        help=('To stop processing '
              'as soon as a syntax error in diagram occurs'),
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-failfast2',
        help=('To do a first syntax check before processing files, '
              'to fail even faster'),
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-pattern',
        help='To print the list of Regular Expression used by PlantUML',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-duration',
        help='To print the duration of complete diagrams processing',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-nbthread',
        help='To use (N) threads for processing',
        action=NotSupportedAction,
        metavar='N')
    parser.add_argument(
        '-timeout',
        help=('Processing timeout in (N) seconds. '
              'Defaults to 15 minutes (900 seconds).'),
        action=NotSupportedAction,
        metavar='N')
    parser.add_argument(
        '-about', '-authors', '-author',
        help='To print information about PlantUML authors',
        action=_FlagAction,
        dest='authors')
    parser.add_argument(
        '-overwrite',
        help='To allow to overwrite read only files',
        action='store_true')
    parser.add_argument(
        '-printfonts',
        help='To print fonts available on your system',
        action=_FlagAction,
        dest='help font')
    parser.add_argument(
        '-enablestats',
        help='To enable statistics computation',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-disablestats',
        help='To disable statistics computation (default)',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-htmlstats',
        help='To output general statistics in file plantuml-stats.html',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-xmlstats',
        help='To output general statistics in file plantuml-stats.xml',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-realtimestats',
        help='To generate statistics on the fly rather than at the end',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-loopstats',
        help='To continuously print statistics about usage',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-dumpstats',
        help='TODO',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-dumphtmlstats',
        help='TODO',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-clipboard',
        help='TODO',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-clipboardloop',
        help='TODO',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-splash',
        help='To display a splash screen with some progress bar',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-progress',
        help='To display a textual progress bar in console',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-pipedelimitor',
        help='TODO',
        action='store')
    parser.add_argument(
        '-pipemap',
        help='Not support currently. TODO',
        action='store_true')
    parser.add_argument(
        '-pipenostderr',
        help='Not support currently. TODO',
        action='store_true')
    parser.add_argument(
        '-pipeimageindex',
        help=('Not support currently. '
              'To generate the Nth image with pipe option'),
        action='store',
        metavar='N')
    parser.add_argument(
        '-stdlib',
        help='To print standard library info',
        action=_FlagAction)
    parser.add_argument(
        '-extractstdlib',
        help='To extract PlantUML Standard Library into stdlib folder',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-stdrpt', '-stdrpt:1',
        help='Not support currently. TODO',
        action='store_true')
    parser.add_argument(
        '-useseparatorminus',
        help='Not support currently. TODO',
        action='store_true')
    parser.add_argument(
        '-filename',
        help='Not support currently. To override %%filename%% variable',
        action='store',
        metavar='"example.puml"')
    parser.add_argument(
        '-preproc',
        help='To output preprocessor text of diagrams',
        action=NotSupportedFlagAction)
    parser.add_argument(
        '-cypher',
        help='To cypher texts of diagrams so that you can share them',
        action=NotSupportedFlagAction)
    parser.add_argument(
        'file/dir',
        help='UML files',
        nargs='*')
    parser.set_defaults(main_function=main)
    return parser


class CompileError(RuntimeError):
    """Exception for compile error.
    """
    def __init__(self, msg, data):
        super(CompileError, self).__init__(msg)
        self.data = data


def encode_code(uml_code, level=-1):
    """Encode PlantUML code by PlantUML Text Encoding
    (http://plantuml.com/en/text-encoding).

    Args:
        uml_code (str): PlantUML code.
        level (int): Compressing level.
    Returns:
        str: The PlantUML Text Encoding text.
    """
    byte_code = uml_code.encode('utf-8')
    compressed_code = zlib.compress(byte_code, level)
    base64_code = base64.b64encode(compressed_code).decode('ascii')
    return ''.join(_ENCODE_TABLE[c] for c in base64_code)


def decode_code(encoded_uml):
    """Decode PlantUML Text Encoding text to PlantUML code.

    Args:
        encoded_uml (str): PlantUML Text Encoding text.
    Returns:
        str: PlantUML code.
    """
    base64_code = ''.join(_DECODE_TABLE[c] for c in encoded_uml)
    compressed_code = base64.b64decode(base64_code)
    wbits = 15 if compressed_code[0:2] == b'\x78\x9c' else -15
    byte_code = zlib.decompress(compressed_code, wbits)
    return byte_code.decode('utf-8')


def compile_code(base_url, uml_code, output_format, use_post=False, timeout=None):
    """Compile PlantUML code into the specified format data by PlantUML server.

    Args:
        base_url (str): URL of PlantUML server.
        uml_code (str): PlantUML code.
        output_format (str): The target format.
        use_post (bool): Whether or not to use HTTP POST method for compiling.
            PlantUML server supports POST method from version 1.2018.5.
        timeout (int): The server communication timeout in seconds.
    Returns:
        bytes: output data with the specified format.
    Raises:
        CompileError: If PlantUML code can not be compiled.
        urllib.error.HTTPError: If some server error occurs.
        urllib.error.URLError: If some protocol error occurs.
    """
    pattern = _FORMAT_TO_URL_PATTERN_TABLE.get(output_format, output_format)
    if use_post:
        url = base_url + '/' + pattern + '/'
        data = uml_code.encode('utf-8')
        headers = {'Content-Type': 'text/plain;charset="UTF-8"'}
        request = urllib.request.Request(url, data, headers)
    else:
        encoded_uml = encode_code(uml_code)
        url = base_url + '/' + pattern + '/' + encoded_uml
        request = urllib.request.Request(url)
    try:
        with urllib.request.urlopen(request, timeout=timeout) as response:
            return response.read()
    except urllib.error.HTTPError as error:
        with error:
            if error.code != 400:
                raise
            raise CompileError(error.reason, error.read())


def get_language(base_url, timeout=None):
    """Get PlantUML languange information.

    Args:
        base_url (str): URL of PlantUML server.
        timeout (int): The server communication timeout in seconds.
    Returns:
        str: PlantUML language information.
    """
    url = base_url + '/language'
    with urllib.request.urlopen(url, timeout=timeout) as response:
        return response.read().decode('utf8')


def format_to_ext(output_format):
    """Get a file extension from the output format

    Args:
        output_format (str): The target format.
    Returns:
        str: The corresponding file extension with '.ext' format.
    """
    if output_format == 'txt':
        return '.atxt'
    if output_format == 'braille':
        return '.braille.png'
    index = output_format.find(':')
    if index != -1:
        return '.' + output_format[0:index]
    return '.' + output_format


def _generate(base_url, timeout,
              filepath, output_format, outdir, output_filepath=None):
    # pylint: disable=too-many-arguments
    uml_code = filepath.read_text()
    try:
        reply = compile_code(base_url, uml_code, output_format, timeout)
        result = True
    except CompileError as error:
        logging.warning('%s: %s', filepath, error)
        reply = error.data
        result = False

    if output_filepath is None:
        output_filepath = filepath.parent / outdir / filepath.name
        output_filepath.parent.mkdir(parents=True, exist_ok=True)
    output_filepath = output_filepath.with_suffix(format_to_ext(output_format))
    output_filepath.write_bytes(reply)
    return result

def _check_syntax(base_url, timeout, filepath):
    uml_code = filepath.read_text()
    try:
        reply = compile_code(
            base_url, uml_code, 'check', use_post=False, timeout=timeout)
        result = True
    except CompileError as error:
        logging.warning('%s: %s', filepath, error)
        reply = error.data
        result = False
    print(filepath, ':', reply.decode('utf-8'))
    return result

def _encodeurl(filepath):
    try:
        uml_code = filepath.read_text()
        print(filepath, ':', encode_code(uml_code))
    except Exception as ex: # pylint: disable=broad-except
        logging.error('Failed to encode %s: %s', uml_code, ex)
        return False
    return True

def _decodeurl(encoded_uml):
    try:
        print(decode_code(encoded_uml))
    except Exception as ex: # pylint: disable=broad-except
        logging.error('Failed to decode %s: %s', encoded_uml, ex)
        return False
    return True

def _for_each_file(paths, proc, do_exit=False):
    result = True
    for path in paths:
        path = os.path.expandvars(os.path.expanduser(path))
        do_process = False
        for filepath in glob.iglob(path, recursive=True):
            filepath = pathlib.Path(filepath)
            if not filepath.is_file():
                continue
            do_process = True
            try:
                result = proc(filepath) and result
            except Exception: # pylint: disable=broad-except
                logging.exception('%s: Failed to process', filepath)
                result = False
        if not do_process:
            logging.warning('%s is invalid path', path)
            continue
    if do_exit:
        sys.exit(0 if result else 1)
    return result

def _read_uml_code():
    code_lines = []
    for line in sys.stdin:
        code_lines.append(line)
        if line.startswith('@end'):
            return '\n'.join(code_lines)
    return '\n'.join(code_lines)

def _run_with_pipe(base_url, timeout, args):
    result = True
    while True:
        uml_code = _read_uml_code()
        if not uml_code:
            break
        try:
            if args.syntax:
                fmt = 'check'
                use_post = False
            else:
                fmt = args.format
                use_post = True
            reply = compile_code(base_url, uml_code, fmt,
                                 use_post=use_post, timeout=timeout)
            sys.stdout.buffer.write(reply)
        except CompileError as error:
            logging.warning('%s', error)
            sys.stdout.buffer.write(error.data)
            result = False
        except Exception: # pylint: disable=broad-except
            logging.exception('Failed to compile')
            result = False
        if args.pipedelimitor is not None:
            print(args.pipedelimitor)
    sys.exit(0 if result else 1)


def main(args):
    """Run action as plantuml command.

    Args:
        args: Parsing result from the parser created by `make_parser`.
    """
    base_url = sabacan.utils.get_server_url('plantuml', DEFAULT_SERVER_URL)
    timeout = sabacan.utils.get_timeout('plantuml')

    if args.pipe:
        _run_with_pipe(base_url, timeout, args)

    input_paths = getattr(args, 'file/dir')
    if args.computeurl:
        _for_each_file(input_paths, _encodeurl, do_exit=True)

    if args.decodeurl:
        result = all(_decodeurl(encoded_uml) for encoded_uml in input_paths)
        sys.exit(0 if result else 1)

    if args.syntax:
        _for_each_file(
            input_paths,
            lambda path: _check_syntax(base_url, timeout, path),
            do_exit=True)

    if args.outfile is not None:
        outfile = pathlib.Path(args.outfile)
        if not outfile.parent.exists(): # TODO argument value check
            logging.error('Invalid output filepath: %s', outfile)
            sys.exit(1)
    else:
        outfile = None
    if args.outdir is not None:
        outdir = pathlib.Path(args.outdir) # TODO argument value check
    else:
        outdir = ''

    _for_each_file(
        input_paths,
        lambda path: _generate(
            base_url, timeout, path, args.format, outdir, outfile),
        do_exit=True)


if __name__ == '__main__':
    ARGS = make_parser().parse_args()
    ARGS.main_function(ARGS)



"""This module provides utility function to sabacan.
"""
import argparse
import logging
import os


class SetEnvAction(argparse.Action): # pylint: disable=too-few-public-methods
    """Custom argparse.Action which set environment variable.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        os.environ[self.dest] = values

def get_server_url(servername, default=None):
    """Get URL for server from environment variables.

    Args:
        servername (str): The name of application (e.g. plantuml)
        default (str): The result when URL done not exist
            in environment variables.
    Returns:
        str: URL of server
    """
    url = os.getenv('SABACAN_%s_URL' % servername.upper())
    if url is not None:
        return url
    return os.getenv('SABACAN_URL', default)

def get_timeout(servername, default=None):
    """Get timeout for server communication from environment variables.

    Args:
        servername (str): The name of application (e.g. plantuml)
        default (int): The result when timeout does not exist
            in environment variables.
    Returns:
        str: The timeout (sec) of server communication.
    """
    timeout = os.getenv('SABACAN_%s_TIMEOUT' % servername.upper())
    if timeout is None:
        timeout = os.getenv('SABACAN_TIMEOUT', default)
    if timeout is not None:
        try:
            return int(timeout)
        except ValueError:
            pass
    return None


class NotSupportedAction(argparse.Action):
    """Custom argparse.Action class for not supported options.

    The help message for an option attached to this class is appended
    'NOT SUPPORTED' to.
    When the options are used, waning messages are displayed automatically.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        # pylint: disable=too-many-arguments,redefined-builtin
        help = 'NOT SUPPORTED' if help is None else help + ' (NOT SUPPORTED)'
        super(NotSupportedAction, self).__init__(
            option_strings, dest, nargs, const,
            default, type, choices, required, help, metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        logging.warning('"%s" is not supported', option_string)


class NotSupportedFlagAction(NotSupportedAction):
    """Custom argparse.Action class for not supported options.

    The class is same as NotSupportedAction except nargs is always 0.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, option_strings, dest, nargs=None, const=None,
                 default=None, type=None, choices=None, required=False,
                 help=None, metavar=None):
        # pylint: disable=too-many-arguments,redefined-builtin
        super(NotSupportedFlagAction, self).__init__(
            option_strings, dest, 0, const,
            default, type, choices, required, help, metavar)



import os
from setuptools import setup

import sabacan

BASEDIR = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(BASEDIR, 'README.rst'), 'r') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='sabacan',
    version=sabacan.__version__,
    description='Command Line Interface for various application servers',
    long_description=LONG_DESCRIPTION,
    url='https://github.com/amedama41/sabacan',
    author='amedama41',
    author_email='kamo.devel41@gmail.com',
    keywords=['plantuml', 'redpen'],
    packages=['sabacan'],
    entry_points={
        'console_scripts': ['sabacan = sabacan:main'],
    },
    python_requires='>=3.5',
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ]
)
